/* -*- C -*-
 * Created: Fri Nov 21 14:35:30 2003

Copyright (C) 2003 Alex Tutubalin, lexa@lexa.ru <p>
      
      , ,  
          ,
      
      <a href="http://www.apache.org/LICENSE.txt">Apache License</a> 
      -         
      <a href="http://www.lexa.ru/lexa/">http://www.lexa.ru/lexa</a>.
           grepip.<p>
            .<p>
*/


#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <pcre.h>

#include "patricia.h"



// params
int buf_size=100;
int column=0;
int print_filenames=1;
int print_unmatching=0;
int pattern_prefix=0;
int pfile_as_prefix=0;
int simple_ips=0;
char *regexp;

// default regexps
static char *default_regex="\\b(\\d{1,3}\\.\\d{1,3}\\.\\d{1,3}\\.\\d{1,3})\\b";
static char *loose_regex="(\\d{1,3}\\.\\d{1,3}\\.\\d{1,3}\\.\\d{1,3})";
static char *space_regex="(\\S+)";


//  global vars
char *regexp=NULL;
char *buf;
pcre_extra *extra=NULL;
pcre *pc=NULL;
patricia_tree_t *list = NULL;
int has_paren;

void add_prefix(char *prefix, char *data)
{
    char sprefix[80];
    patricia_node_t *node=NULL;
    prefix_t pref;
    pref.family = AF_INET;
    pref.ref_count = 0;
    if(simple_ips && !strchr(prefix,'/'))
	{
	    int count=0;
	    char *p=prefix;
	    if(!prefix||strlen(prefix)<1)
		return;
	    while((p=strchr(p,'.')))
		{
		    p++;
		    count++;
		}
	    if(count>2)
		{
		    pref.bitlen = 32;
		    if(inet_aton(prefix,&pref.add.sin))
			node = patricia_lookup (list, &pref);
		}
	    else if(count==2)
		{
		    pref.bitlen = 24;
		    snprintf(sprefix,80,"%s.0",prefix);
		    if(inet_aton(sprefix,&pref.add.sin))
			node = patricia_lookup (list, &pref);
		}
	    else if (count == 1)
		{
		    pref.bitlen = 16;
		    snprintf(sprefix,80,"%s.0.0",prefix);
		    if(inet_aton(sprefix,&pref.add.sin))
			node = patricia_lookup (list, &pref);
		}
	    else
		{
		    pref.bitlen = 8;
		    snprintf(sprefix,80,"%s.0.0.0",prefix);
		    if(inet_aton(sprefix,&pref.add.sin))
			node = patricia_lookup (list, &pref);
		}
	    if(node)
		node->data = data?data:"";
	    return;
	}
    if(prefix && strlen(prefix)>2)
	{
	    node = make_and_lookup(list,prefix);
	    if(node)
		node->data = data?data:"";
	}
}

void add_string_prefix(char *buf,char *filename)
{
    char *p,*bp;
    bp = buf;
    if((p=strchr(bp,'#'))) *p=0;
    while(*bp && isspace(*bp)) bp++;
    if(!*bp) return;
    if(!pfile_as_prefix)
	{
	    if((p=strchr(bp,'\n'))) *p=0;
	    if((p=strchr(bp,'\r'))) *p=0;
	    p=strchr(bp,' ');
	    if(!p)
		p=strchr(bp,'\t');
	    if(p)
		{
		    *p=0;
		    p++;
		}
	}
    add_prefix(bp,pfile_as_prefix?filename:(p?strdup(p):NULL));
}

void add_file(char *filename)
{
    FILE *f = fopen(filename,"r");
    char buf[5120];
    char *prefix = strdup(filename);
    if(!f) return;
    while(fgets(buf,5120,f))
	{
	    add_string_prefix(buf,prefix);
	}
    fclose(f);
}

void usage(void)
{
    fprintf(stderr,"Usage: ipgrep [options] PATTERN file1 file2\n");
    fprintf(stderr,"\tPATTERN - IP range(prefix/len) + optional tag, i.e. '192.168.0.0/24 PVT'\n");
    fprintf(stderr,"Output options:\n");
    fprintf(stderr,"\t-v\t- print unmatching lines\n");
    fprintf(stderr,"\t-h\t- supress prefix printing\n");
    fprintf(stderr,"\t-p\t- print pattern tag instead of filenames\n");
    fprintf(stderr,"\t-P\t- use matching pattern filename as prefix (use with -f)\n");
    fprintf(stderr,"\t-d\t- dump pattern list and exit\n");
    fprintf(stderr,"Patterns options:\n");
    fprintf(stderr,"\t-f filename - read patterns from filename (PATTERN argument unused)\n");
    fprintf(stderr,"\t-e PATTERN  - specify additional pattern\n");
    fprintf(stderr,"\t-s\t- use simple network patterns, so x.y.z will\n");
    fprintf(stderr,"\t\t  become x.y.z.0/24 and so on\n");
    fprintf(stderr,"Input matching options:\n");
    fprintf(stderr,"\t-R regexp - use regexp (Perl style) to find IPs in input files.\n");
    fprintf(stderr,    "\t\t    Use (  ) to separate IP within pattern.\n");
    fprintf(stderr,    "\t\t    Default regexp: \'\\b(\\d{1,3}\\.\\d{1,3}\\.\\d{1,3}\\.\\d{1,3})\\b\'\n");
    fprintf(stderr,"\t-l\t- use \'loose\' regexp  \'(\\d{1,3}\\.\\d{1,3}\\.\\d{1,3}\\.\\d{1,3})\'\n");
    fprintf(stderr,"\t-w\t- use \'(\\S+)\' regex - split input into fields by spaces\n");
    fprintf(stderr,"\t-c N\t- check only Nth column (with -w)/Nth occurence of regex\n");
    fprintf(stderr,"Misc options:\n");
    fprintf(stderr,"\t-b kbytes   - maximum expected stling length(default 100kb)\n");
    
    exit(0);
}

void print_string(char *string, char *matchstring, char *filestring)
{
    if(print_unmatching)
	{
	    if(print_filenames && !pattern_prefix && filestring)
		printf("%s: %s",filestring,string);
	    else 
		printf("%s",string);
	}
    else
	{
	    if(!print_filenames 
	       || (pattern_prefix && !matchstring)
	       || !filestring) // no suitable prefix
		printf("%s",string);
	    else if(pattern_prefix && matchstring)
		printf("%s: %s",matchstring,string);
	    else
		printf("%s: %s",filestring,string);
	}
}

char *match_string(char *buf, size_t len)
{
    int col=0,rc,i;
    char ipbuf[20];
    int ofs = 0;
    int ovector[30];
    while((rc = pcre_exec(pc,extra,buf,len,ofs,PCRE_NOTEMPTY,ovector,30))>0)
	{
	    int len;
	    ofs=ovector[1];
	    col++;
	    if(column && col!=column)
		continue;
	    for(i=has_paren;i<rc;i++)
		{
		    if((len=pcre_copy_substring(buf,ovector,rc,i,ipbuf,20))>0 && len < 20)
			{
			    prefix_t prefix;
			    char *p=ipbuf; 
			    int cnt = 0;
			    // count dots
			    while((p=strchr(p,'.')))
			     {
			      p++; cnt++;
			     }
			    if(cnt!=3) continue;
			    prefix.family = AF_INET;
			    prefix.bitlen = 32;
			    prefix.ref_count = 0;
			    if(inet_aton(ipbuf,&prefix.add.sin))
				{
				    patricia_node_t *node = patricia_search_best(list,&prefix);
				    if(node)
					return node->data;
				}

			}
		}
	    ofs=ovector[1];
	}
    return NULL;
}

void process_file(FILE*f, char *p) 
{
    char  *match;
    if(!buf) return;
    while(fgets(buf,buf_size*1024,f))
	{
	    match = match_string(buf,strlen(buf));
	    if ((match && !print_unmatching)
		||
		(!match && print_unmatching))
		print_string(buf,match,p);
	}
}


void dump_patterns_cb(void *p, void *d)
{
    prefix_t *pp = (prefix_t*)p;
    char *s=(char*)d;
    printf("%s/%d %s\n",inet_ntoa(pp->add.sin),pp->bitlen,s?s:"");

}

int pcre_prepare(char *regex,pcre **pcre,pcre_extra **extra, const char **errptr, int *erroff)
{
    *errptr=NULL;
    *erroff=0;
    has_paren=strchr(regex,'(')?1:0;
    *pcre=pcre_compile(regex,0,errptr,erroff,NULL);
    if(*pcre)
	{
	    *extra = pcre_study(*pcre,0,errptr);
	    if(!*extra && *errptr)
		return -2;
	    return 0;
	}
    else
	return -1;
}

int main(int argc, char *argv[])
{
    int patterns_loaded = 0,i,dump=0;
    const char *errptr;
    int  erroff;
    list = New_Patricia(32);
    if(!list)
	{
	    perror("alloc");
	    exit(1);
	}
    regexp = default_regex;
    while((i=getopt(argc,argv,"vhpPdf:e:sR:lwc:b:")) != EOF)
	{
	    switch(i)
		{
// Output options
		case 'v': print_unmatching=1; break;
		case 'h': print_filenames=0; break;
		case 'p': pattern_prefix = 1; break;
		case 'P': pfile_as_prefix = 1; break;
		case 'd': dump = 1; break;
// Patterns options
		case 'f': add_file(optarg); patterns_loaded = 1; break;
		case 'e': add_string_prefix(optarg,NULL); break;
		case 's': simple_ips = 1; break;
// Input matching
		case 'R': regexp = strdup(optarg); break;
		case 'l': regexp = loose_regex;   break;
		case 'w': regexp = space_regex; break;
		case 'c': column = atoi(optarg); break;
// Misc options
		case 'b': buf_size=atoi(optarg)>buf_size?atoi(optarg):buf_size; break;
		default:  usage(); break;
		}
	}
    argc-=optind;
    argv+=optind;
    if(!patterns_loaded)
	{
	    if(argc>0)
		{
		    add_string_prefix(argv[0],NULL);
		    argc--;
		    argv++;
		}
	    else
		usage();
	}
    if(dump)
	{
	    patricia_process(list,dump_patterns_cb);
	    exit(0);
	}

    switch(pcre_prepare(regexp,&pc,&extra,&errptr,&erroff))
	{
	case -1:
	    fprintf(stderr,"Cannot compile regex %s at point %s: %s\n",regexp,regexp+erroff,errptr);
	    exit(1);
	    break;
	case -2:
	    fprintf(stderr,"Cannot optimize regex %s: %s\n",regexp,errptr);
	    exit(0);
	    break;
	default:
	    break;
	}
    if(!(buf = malloc(buf_size*1024)))
	{
	    perror("malloc");
	    exit(1);
	}
    if(argc<2 && !pattern_prefix)
	print_filenames = 0;
    if(argc==0)
	process_file(stdin,NULL);
    else
	for(i=0;i<argc;i++)
	    {
		FILE *f = fopen(argv[i],"r");
		if(f) {process_file(f,argv[i]); fclose(f);}
		else perror(argv[i]);
	    }
	    
    return 0;
}

